/* Code to read & store MORAY data files */

/* Platform/compiler specific include files */
#if defined( MAC )
#include <console.h>
#include <profile.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "mryread.h"

typedef float Vec[3];
typedef float Matrix[4][4];

#define EPSILON 1.0e-10
#if !defined( M_PI )
#define M_PI 3.1415926535897932384626
#endif
#define degtorad(x)     (((double)(x))*M_PI/180.0)
#define radtodeg(x)     (((double)(x))*180.0/M_PI)
#define MakeVector(x, y, z, v) {(v)[0]=(x);(v)[1]=(y);(v)[2]=(z);}
#define VecScale(S,a)   {(a)[0] *= S ; (a)[1] *= S ; (a)[2] *= S;}
#define VecDot(a,b)   ((a)[0]*(b)[0]+(a)[1]*(b)[1]+(a)[2]*(b)[2])
#define VecLen(a)     (sqrt(VecDot(a,a)))
#define VecCopy(a,b)    {(b)[0]=(a)[0];(b)[1]=(a)[1];(b)[2]=(a)[2];}
#define VecAdd(a,b,c)    {(c)[0]=(a)[0]+(b)[0];\
                          (c)[1]=(a)[1]+(b)[1];\
                          (c)[2]=(a)[2]+(b)[2];}
#define VecSub(a,b,c)    {(c)[0]=(a)[0]-(b)[0];\
                          (c)[1]=(a)[1]-(b)[1];\
                          (c)[2]=(a)[2]-(b)[2];}
#define VecComb(A,a,B,b,c) {(c)[0]=(A)*(a)[0]+(B)*(b)[0];\
                            (c)[1]=(A)*(a)[1]+(B)*(b)[1];\
                            (c)[2]=(A)*(a)[2]+(B)*(b)[2];}
#define VecCross(a,b,c)    {(c)[0]=(a)[1]*(b)[2]-(a)[2]*(b)[1];\
                            (c)[1]=(a)[2]*(b)[0]-(a)[0]*(b)[2];\
                            (c)[2]=(a)[0]*(b)[1]-(a)[1]*(b)[0];}

typedef struct tx_struct *tx_ptr;
typedef struct tx_struct {
   Matrix tx;
   tx_ptr next;
   };

static Matrix CurrentTx =
   {{1, 0, 0, 0},
    {0, 1, 0, 0},
    {0, 0, 1, 0},
    {0, 0, 0, 1}};
static tx_ptr TxStack = NULL;

static void
lib_zero_matrix(Matrix mx)
{
    int i, j;

    for (i=0;i<4;++i)
        for (j=0;j<4;++j)
            mx[i][j] = 0.0;
}

/* Create identity matrix. */
static void
lib_create_identity_matrix(Matrix mx)
{
    int i;

    lib_zero_matrix(mx);
    for (i=0;i<4;++i)
        mx[i][i] = 1.0;
}

/* Copy one matrix to another */
static void
lib_copy_matrix(Matrix mres, Matrix mx)
{
    int i, j;

    for (i=0;i<4;i++)
        for (j=0;j<4;j++)
            mres[i][j] = mx[i][j];
}

/* Create translation matrix */
static void
lib_create_translate_matrix(Matrix mx, Vec vec)
{
    lib_create_identity_matrix(mx);
    mx[3][0] = vec[0];
    mx[3][1] = vec[1];
    mx[3][2] = vec[2];
}

/* Create scaling matrix */
static void
lib_create_scale_matrix(Matrix mx, Vec vec)
{
    lib_zero_matrix(mx);
    mx[0][0] = vec[0];
    mx[1][1] = vec[1];
    mx[2][2] = vec[2];
    mx[3][3] = 1.0;
}

static void
lib_create_rotate_matrix(Matrix mx, int axis, float angle)
{
    double cosine, sine;

    lib_zero_matrix(mx);
    cosine = cos((double)angle);
    sine = sin((double)angle);
    switch (axis) {
        case 0:
            mx[0][0] = 1.0;
            mx[1][1] = cosine;
            mx[2][2] = cosine;
            mx[1][2] = sine;
            mx[2][1] = -sine;
            break ;
        case 1:
            mx[1][1] = 1.0;
            mx[0][0] = cosine;
            mx[2][2] = cosine;
            mx[2][0] = sine;
            mx[0][2] = -sine;
            break;
        case 2:
            mx[2][2] = 1.0;
            mx[0][0] = cosine;
            mx[1][1] = cosine;
            mx[0][1] = sine;
            mx[1][0] = -sine;
            break;
    }
    mx[3][3] = 1.0;
}

/* Multiply a direction by a matrix. */
static void
lib_transform_vector(Vec vres, Vec vec, Matrix mx)
{
   Vec vtemp;

   vtemp[0] = vec[0]*mx[0][0] + vec[1]*mx[1][0] + vec[2]*mx[2][0];
   vtemp[1] = vec[0]*mx[0][1] + vec[1]*mx[1][1] + vec[2]*mx[2][1];
   vtemp[2] = vec[0]*mx[0][2] + vec[1]*mx[1][2] + vec[2]*mx[2][2];

   VecCopy(vtemp, vres);
}

/* Multiply a point by a matrix. */
static void
lib_transform_point(Vec vres, Vec vec, Matrix mx)
{
   Vec vtemp;

   vtemp[0] = vec[0]*mx[0][0] + vec[1]*mx[1][0] + vec[2]*mx[2][0] + mx[3][0];
   vtemp[1] = vec[0]*mx[0][1] + vec[1]*mx[1][1] + vec[2]*mx[2][1] + mx[3][1];
   vtemp[2] = vec[0]*mx[0][2] + vec[1]*mx[1][2] + vec[2]*mx[2][2] + mx[3][2];

   VecCopy(vtemp, vres);
}

/* Multiply two 4x4 matrices. */
static void
lib_matrix_multiply(Matrix mxres, Matrix mx1, Matrix mx2)
{
    int i, j;
    Matrix mtemp;

    for (i=0;i<4;i++)
        for (j=0;j<4;j++)
            mtemp[i][j] = mx1[i][0]*mx2[0][j] + mx1[i][1]*mx2[1][j] +
                          mx1[i][2]*mx2[2][j] + mx1[i][3]*mx2[3][j];
    memcpy(mxres, mtemp, sizeof(Matrix));
}

/* Return 1 if there is an active transformation, 0 if not */
static int
lib_tx_active()
{
   int i, j;
   for (i=0;i<4;i++)
      for (j=0;j<4;j++)
         if (fabs(CurrentTx[i][j] - (i == j ? 1.0 : 0.0)) > EPSILON)
            return 1;
   return 0;
}

/* Return 1 if there is an active transformation, 0 if not */
static void
lib_get_current_tx(Matrix mat)
{
   memcpy(mat, CurrentTx, sizeof(Matrix));
}

static void
lib_tx_pop()
{
   tx_ptr last_tx;

   if (TxStack == NULL) {
      fprintf(stderr, "Attempt to pop beyond bottom of transform stack\n");
      }
   else {
      last_tx = TxStack;
      lib_copy_matrix(CurrentTx, last_tx->tx);
      TxStack = TxStack->next;
      free(last_tx);
      }
}

static void
lib_tx_push()
{
   tx_ptr new_tx;

    if ((new_tx = malloc(sizeof(struct tx_struct))) == NULL)
        error("Failed to allocate polygon data\n");
    lib_copy_matrix(new_tx->tx, CurrentTx);
    new_tx->next = TxStack;
    TxStack = new_tx;
}

static void
lib_tx_rotate(int axis, float angle)
{
    Matrix mx1, mx2;

    lib_create_rotate_matrix(mx1, axis, angle);
    lib_matrix_multiply(CurrentTx, mx1, CurrentTx);
}

static void
lib_tx_scale(Vec vec)
{
    Matrix mx1, mx2;

    lib_create_scale_matrix(mx1, vec);
    lib_matrix_multiply(CurrentTx, mx1, CurrentTx);
}

static void
lib_tx_translate(Vec vec)
{
    Matrix mx1, mx2;

    lib_create_translate_matrix(mx1, vec);
    lib_matrix_multiply(CurrentTx, mx1, CurrentTx);
}

static int
test_transform_values(MORAY_Object *obj)
{
   if (fabs(obj->base_data.sx - 1.0) > EPSILON ||
       fabs(obj->base_data.sy - 1.0) > EPSILON ||
       fabs(obj->base_data.sz - 1.0) > EPSILON ||
       fabs(obj->base_data.rx) > EPSILON ||
       fabs(obj->base_data.ry) > EPSILON ||
       fabs(obj->base_data.rz) > EPSILON ||
       fabs(obj->base_data.tx) > EPSILON ||
       fabs(obj->base_data.ty) > EPSILON ||
       fabs(obj->base_data.tz) > EPSILON)
      return 1;
   else
      return 0;
}

static void
write_transform_info(MORAY_Object *obj)
{
   /* Transformation information */
   if (fabs(obj->base_data.sx - 1.0) > EPSILON ||
       fabs(obj->base_data.sy - 1.0) > EPSILON ||
       fabs(obj->base_data.sz - 1.0) > EPSILON)
      message("scale <%g,%g,%g> ",
              obj->base_data.sx,
              obj->base_data.sy,
              obj->base_data.sz);

   if (fabs(obj->base_data.rx) > EPSILON ||
       fabs(obj->base_data.ry) > EPSILON ||
       fabs(obj->base_data.rz) > EPSILON)
      message("rotate <%g,%g,%g> ",
              -obj->base_data.rx,
              -obj->base_data.ry,
              -obj->base_data.rz);

   if (fabs(obj->base_data.tx) > EPSILON ||
       fabs(obj->base_data.ty) > EPSILON ||
       fabs(obj->base_data.tz) > EPSILON)
      message("translate <%g,%g,%g> ",
              obj->base_data.tx,
              obj->base_data.ty,
              obj->base_data.tz);
}

static float
Convert_Phong(float phong_pow)
{
   if (phong_pow <= 0.0)
      phong_pow = 1.0e-5;
   else if (phong_pow < 1000.0)
      phong_pow = 180.0 * acos(exp(log(2.0) / -phong_pow)) / (2.0 * 3.14159);

   return phong_pow;
}

static void
write_surface_components(MRY_Texture_Data *tex)
{
   if (tex->amb != 0.0)
      message("   ambient %g\n", tex->amb);
   if (tex->dif != 0.0)
      message("   diffuse %g\n", tex->dif);
   if (tex->bril != 1.0)
      message("   brilliance %g\n", tex->bril);
   if (tex->refl != 0.0)
      message("   reflection white, %g\n", tex->refl);
   if (tex->trns != 0.0)
      message("   transmission white, %g, %g\n", tex->trns, tex->rfin);
   if (tex->phg != 0.0) {
      message("   specular <%g,%g,%g>, %g\n",
              (double)tex->clsr, (double)tex->clsg,
              (double)tex->clsb, (double)tex->phg);
      message("   microfacet Phong %g\n", Convert_Phong(tex->phgx));
      }
   if (tex->spec != 0.0) {
      message("   specular <%g,%g,%g>, %g\n",
              tex->clsr, tex->clsg, tex->clsb, tex->phg);
      /* The conversion to an angle for POV-Ray's roughness is just
         a guess.  Need to actually determine the formula. */
      if (tex->rghn < 1.0e-10)
         message("   microfacet Cook 0.1\n");
      else
         message("   microfacet Cook %g\n", Convert_Phong(1.0 / tex->rghn));
      }
}

static void
Declare_Textures(MORAY_Object *textures)
{
   static int include_file_flag = 0;
   int i;
   MORAY_Object *texs;
   MRY_Texture_Data *tex;
   MRY_Checker_Data *checker;
   MRY_Imagemap_Data *image;

   for (texs=textures;texs!=NULL;texs=texs->next) {
      tex = texs->obj_data;

      /* Correct any spaces in the middle of the texture name */
      for (i=0;i<MORAY_OBJ_NAME_LEN && texs->base_data.name[i] != '\0';i++)
         if (texs->base_data.name[i] == ' ')
            texs->base_data.name[i] = '_';

      /* Output the texture definition */
      if (tex->texture_type == MORAY_OPAQUE_TEXTURE) {
         message("define %s texture { surface {\n", &texs->base_data.name[0]);
         message("   color <%g,%g,%g>\n", tex->cldr, tex->cldg, tex->cldb);
         write_surface_components(tex);
         message("   } }\n");
         }
      else if (tex->texture_type == MORAY_PREDEFINED_TEXTURE) {
	 if (!include_file_flag) {
	    message("include \"povtex.inc\"\n");
	    include_file_flag = 1;
	 }
         message("// Uses predefined texture: %s\n",
                 texs->base_data.name);
         }
      else if (tex->texture_type == MORAY_CHECKER_TEXTURE) {
         checker = (MRY_Checker_Data *)tex->unique_data;
         message("define %s texture { checker\n", &texs->base_data.name[0]);
         for (i=0;i<2;i++) {
            if (i==0) {
               message(" texture { surface {\n");
               message("   color <%g,%g,%g>\n",
                       tex->cldr, tex->cldg, tex->cldb);
               write_surface_components(tex);
               message("   } },\n");
               }
            else {
               message(" texture { surface {\n");
               message("   color <%g,%g,%g>\n",
                       checker->cl2r, checker->cl2r, checker->cl2r);
               write_surface_components(tex);
               message("   } }\n");
               }
            }
         if (test_transform_values(texs)) {
            message("   ");
            write_transform_info(texs);
            message("\n");
            }
         message(" }\n");
         }
      else if (tex->texture_type == MORAY_IMAGEMAP_TEXTURE) {
         image = (MRY_Imagemap_Data *)tex->unique_data;
         message("define %s texture { special surface {\n",
                 &texs->base_data.name[0]);
         message("   color ");
         if (image->mapt == MORAY_PLANAR_IMAGEMAP)
            message("planar_imagemap");
         else if (image->mapt == MORAY_SPHERICAL_IMAGEMAP)
            message("spherical_imagemap");
         else if (image->mapt == MORAY_CYLINDRICAL_IMAGEMAP)
            message("cylindrical_imagemap");
         else
            message("planar_imagemap"); /* Can't handle this type... */
         message("(image(\"%s\"), <x,z,y>%s)\n", image->name,
                 (image->tiled ? ",1" : ""));
         write_surface_components(tex);
         message(" } ");
         if (test_transform_values(texs))
            write_transform_info(texs);
         message(" }\n");
         }
      else if (tex->texture_type == MORAY_GRADIENT_TEXTURE) {
         message("define %s texture { special surface {\n",
                 &texs->base_data.name[0]);
         message("   color color_map([0, 1, <%g,%g,%g>, white])[fmod(x, 1)]\n",
                 tex->cldr, tex->cldg, tex->cldb);
         write_surface_components(tex);
         message(" } ");
         if (test_transform_values(texs))
            write_transform_info(texs);
         message(" }\n");
         }
      else if (tex->texture_type == MORAY_WOOD_TEXTURE) {
         message("define %s texture { special surface {\n", &texs->base_data.name[0]);
         message("   color color_map([0, 1, <%g,%g,%g>, white])",
                 tex->cldr, tex->cldg, tex->cldb);
         message("[sawtooth(sqrt(x^2+z^2))]\n");
         write_surface_components(tex);
         message(" } ");
         if (test_transform_values(texs))
            write_transform_info(texs);
         message(" }\n");
         }
      else
         message("Can't handle texture type: %d\n", tex->texture_type);
      }
}

static void
indent(int depth)
{
   int i;
   for (i=0;i<depth;i++)
      message("   ");
}

static void
write_texture(int depth, MORAY_Object *obj)
{
   /* Surface texturing information */
   if (obj->texture != NULL) {
      indent(depth);
      message("%s\n", obj->texture->base_data.name);
      }
}

static int
modify_sweep_points(int npoints, MRY_Sweep_Point *points, double (*newpts)[3])
{
   int i, j, n;
   double t, d;
   double x0, y0, xt0, yt0;
   double x1, y1, xt1, yt1;

   x0  = points[0].xpos;
   y0  = points[0].ypos;
   xt0 = cos(degtorad(-points[0].angl2));
   yt0 = sin(degtorad(-points[0].angl2));
   n = 0;

   newpts[n][0] = x0;
   newpts[n][1] = y0;
   newpts[n][2] = 0.0;
   n++;

   for (i=1;i<=npoints;i++) {
      j = i % npoints;
      /* Put previous end point onto the list as the current start point.
         Note that duplicates are going to happen, this is ok since we
         need to be able to handle varying tangents going in and out of
         the original vertices. (This also takes care of putting the
         first vertex on the list.) */

      /* Work with angles between last point and this one */
      x1  = points[j].xpos;
      y1  = points[j].ypos;
      xt1 = cos(degtorad(-points[j].angl1));
      yt1 = sin(degtorad(-points[j].angl1));

      if (xt0 * xt1 + yt0 * yt1 < 0.0) {
         /* Tangents point away from each other, correct */
         xt0 *= -1.0;
         yt0 *= -1.0;
         }

      /* Calculate and add the floating intermediate point */
      d = sqrt((x1-x0)*(x1-x0)+(y1-y0)*(y1-y0));
      if (fabs(xt0) < EPSILON) {
         /* Can't eliminate yt0 */
         if (fabs(xt1) < EPSILON || fabs(yt0) < EPSILON)
            /* Degenerate, the lines don't intersect */
            goto next_point;
         t = (x0 - x1) / xt1;
         if (fabs(t) < EPSILON || fabs(t-d) < EPSILON)
            goto next_point;
         newpts[n][0] = x1 + t * xt1;
         newpts[n][1] = y1 + t * yt1;
         newpts[n][2] = 1.0;
         n++;
         }
      else if (fabs(yt1) < EPSILON) {
         if (fabs(yt0) < EPSILON)
            goto next_point;
         t = (y0 - y1) / yt0;
         if (fabs(t) < EPSILON || fabs(t-d) < EPSILON)
            goto next_point;
         newpts[n][0] = x0 + t * xt0;
         newpts[n][1] = y0 + t * yt0;
         newpts[n][2] = 1.0;
         n++;
         }
      else if (fabs(xt1 - xt0) < EPSILON || fabs(xt1 + xt0) < EPSILON)
         goto next_point;
      else {
         t = -yt0 / xt0;
         t = (t * (x0 - x1) + (y0 - y1)) / (t * xt1 + yt1);
         if (fabs(t) < EPSILON || fabs(t-d) < EPSILON)
            goto next_point;
         newpts[n][0] = x1 + t * xt1;
         newpts[n][1] = y1 + t * yt1;
         newpts[n][2] = 1.0;
         n++;
         }

next_point:
      /* Always add the next vertex to the list */
      newpts[n][0] = x1;
      newpts[n][1] = y1;
      newpts[n][2] = 0.0;
      n++;

      /* Set up for the next point */
      x0  = x1;
      y0  = y1;
      xt0 = cos(degtorad(-points[j].angl2));
      yt0 = sin(degtorad(-points[j].angl2));
      }

  return n;
}

static void
dump_lathe(int depth, MRY_Sweep_Data *sweep)
{
   int i, j, npoints;
   double (*newpts)[3];

   /* Allocate memory for conversion of lathe vertex coordinates */
   newpts = malloc((3 * sweep->numseg) * 3 * sizeof(double));
   if (newpts == NULL) {
      fprintf(stderr, "Out of memory\n");
      exit(1);
      }

   /* Convert from Moray lathe representation to Polyray lathe representation */
   npoints = modify_sweep_points(sweep->numseg, sweep->points, newpts);

   indent(depth);
   message("lathe 3, <0, 0, 1>, %d,\n", npoints);
   indent(depth);
   for (i=0;i<npoints;i++) {
      message("<%g,%g%s>", newpts[i][0], newpts[i][1],
              (newpts[i][2] != 0.0 ? ",1" : ""));
      if (i < npoints-1)
         message(", ");
      else
         message("\n");
      }

   /* Free up dynamic memory */
   free(newpts);
}

static void
dump_glyph(int depth, MRY_Sweep_Data *sweep)
{
   int i, j, npoints;
   double (*newpts)[3];

   /* Allocate memory for conversion of lathe vertex coordinates */
   newpts = malloc((3 * sweep->numseg) * 3 * sizeof(double));
   if (newpts == NULL) {
      fprintf(stderr, "Out of memory\n");
      exit(1);
      }

   /* Convert from Moray lathe representation to Polyray lathe representation */
   npoints = modify_sweep_points(sweep->numseg, sweep->points, newpts);

   indent(depth);
   message("glyph 1 contour %d,\n", npoints);
   indent(depth);
   for (i=0;i<npoints;i++) {
      message("<%g,%g%s>", newpts[i][0], newpts[i][1],
              (newpts[i][2] != 0.0 ? ",1" : ""));
      if (i < npoints-1)
         message(", ");
      else
         message("\n");
      }

   /* Free up dynamic memory */
   free(newpts);
}

static void
dump_bezier(int depth, MRY_Bezier_Data *bezier)
{
   int i, j, k, n;
   MRY_Bezier_Patches *patches, *patch;
   MRY_Bezier_Vertices *points;

   n = bezier->xptch * bezier->yptch;
   points = bezier->points;
   patches = bezier->patches;
   for (i=0,patch=patches;i<n;i++,patch++) {
      indent(depth);
      if (i > 0)
         message("+");
      message("object { bezier 2,0.0001,%d,%d,",
              (int)patch->subdiv, (int)patch->subdiv);
      for (j=0;j<16;j++) {
         message("<%g,%g,%g>",
                 points[patch->vindices[j]].xpos,
                 points[patch->vindices[j]].ypos,
                 points[patch->vindices[j]].zpos);
         if (j<15)
            message(",");
         }
      message("}\n");
      }
}

static void
dump_heightfield(int depth, MORAY_Object *obj)
{
   MRY_Heightfield_Data *height;

   height = (MRY_Heightfield_Data *)obj->obj_data;
   indent(depth);
   message("  object { %s \"%s\" ",
           (height->smooth ? "smooth_height_field" : "height_field" ),
           height->name);
   message("scale <2,-1/255,2> translate <-1,-0.5,-1> rotate <-90, 0, 0> }\n");
   indent(depth);
   message("& object { box <-1,-1,%g>,<1,1,1> }\n", height->water_level);
}

static void
dump_disc(int depth, MORAY_Object *obj)
{
   MRY_Disc_Data *disc = obj->obj_data;

   indent(depth);
   if (disc->inner_radius == 0.0)
      message("disc <0,0,1>,<0,0,1>,%g\n", disc->outer_radius);
   else
      message("disc <0,0,1>,<0,0,1>,%g\n",
              disc->inner_radius, disc->outer_radius);
}

static void
dump_torus(int depth, MORAY_Object *obj)
{
   MRY_Torus_Data *torus;

   torus = (MRY_Torus_Data *)obj->obj_data;
   indent(depth);
   message("torus %g, %g, <%g,%g,%g>, <%g,%g,%g>\n",
           0.5 * (torus->outer_radius + torus->inner_radius),
           0.5 * (torus->outer_radius - torus->inner_radius),
           0.0, 0.0, 0.0, 0.0, 0.0, 1.0);
}

static void
dump_plane(int depth)
{
   indent(depth);
   message("polynomial z\n");
}

static void
dump_cylinder(int depth)
{
   indent(depth);
   message("  object { cylinder <0,0,-0.001>,<0,0,1.001>,1 }\n");
   indent(depth);
   message("* object { disc <0,0,0>,<0,0,-1>,1.001 }\n");
   indent(depth);
   message("* object { disc <0,0,1>,<0,0,1>,1.001 }\n");
}

static void
dump_cone(int depth)
{
   indent(depth);
   message("  object { cone <0,0,0>,0,<0,0,1.001>,1 }\n");
   indent(depth);
   message("* object { disc <0,0,1>,<0,0,1>,1.001 }\n");
}

static void
dump_camera(MORAY_Model *model, MRY_Camera_Data *camera)
{

   message("viewpoint {\n");
   message("   from <%g,%g,%g>\n",
           (double)camera->eyepx,
           (double)camera->eyepy,
           (double)camera->eyepz);
   message("   at <%g,%g,%g>\n",
           (double)camera->eyelx,
           (double)camera->eyely,
           (double)camera->eyelz);
   message("   up <0, 0, 1>\n"); /* MORAY assumes z-axis is up */
   message("   aspect %g\n", (double)-model->Aspect);
   message("   angle %g\n", (double)camera->aprtr);
   message("   resolution %d,%d\n",
           (int)((float)model->XResolution * model->Aspect),
           model->XResolution);
   message("   }\n");
}

static void
dump_light(int depth, MORAY_Object *obj)
{
   indent(depth);
   if (obj->type == MORAY_PLIGHT_TYPE) {
      message("light <%g, %g, %g>, <0, 0, 0>\n",
              ((MRY_Point_Light_Data *)obj->obj_data)->color[0],
              ((MRY_Point_Light_Data *)obj->obj_data)->color[1],
              ((MRY_Point_Light_Data *)obj->obj_data)->color[2]);
      }
   else if (obj->type == MORAY_SLIGHT_TYPE) {
      message("spot_light <%g,%g,%g>,<0,0,0>,<0,0,-1>,%g,%g,%g\n",
              ((MRY_Spot_Light_Data *)obj->obj_data)->color[0],
              ((MRY_Spot_Light_Data *)obj->obj_data)->color[1],
              ((MRY_Spot_Light_Data *)obj->obj_data)->color[2],
              (double)((MRY_Spot_Light_Data *)obj->obj_data)->tightness,
              (double)((MRY_Spot_Light_Data *)obj->obj_data)->radius,
              (double)((MRY_Spot_Light_Data *)obj->obj_data)->falloff);
      }
   else if (obj->type == MORAY_ALIGHT_TYPE) {
      message("object { textured_light { color ");
      message("<%g,%g,%g> polygon %d,%d,%d,%g } }\n",
              ((MRY_Area_Light_Data *)obj->obj_data)->color[0],
              ((MRY_Area_Light_Data *)obj->obj_data)->color[1],
              ((MRY_Area_Light_Data *)obj->obj_data)->color[2],
              (int)((MRY_Area_Light_Data *)obj->obj_data)->xres,
              (int)((MRY_Area_Light_Data *)obj->obj_data)->yres,
              (int)((MRY_Area_Light_Data *)obj->obj_data)->adaptive,
              (((MRY_Area_Light_Data *)obj->obj_data)->jittr ? 0.001 : 0.0));
      indent(depth);
      message("+ object { polygon 4, <0,0,0>,<0,0,1>,<1,0,1>,<1,0,0>");
      message(" shading_flags 0\n");
      indent(depth+1);
      message("texture { surface { color <%g,%g,%g> ambient 1 diffuse 0 }}}\n",
              ((MRY_Area_Light_Data *)obj->obj_data)->color[0],
              ((MRY_Area_Light_Data *)obj->obj_data)->color[1],
              ((MRY_Area_Light_Data *)obj->obj_data)->color[2]);
      indent(depth);
      message("translate <-0.5,0,-0.5> scale <2,1,2> rotate <90,0,0>\n");
      }
}

static void
Declare_Camera(MORAY_Model *model)
{
   MORAY_Object *obj;
   int camera_flag;

   /* We only declare the first camera we find */
   for (obj=model->Camera_List,camera_flag=1;
        obj!=NULL && camera_flag;
        obj=obj->next)
      if (obj->type == MORAY_CAMERA_TYPE) {
         /* Declare the viewpoint */
         camera_flag = 0;
         dump_camera(model, obj->obj_data);
         }

   /* If there wasn't a camera then say so. */
   if (camera_flag)
      message("// No declared camera\n");
}

static void
Declare_Objects(MORAY_Object *objects, int depth, char seper)
{
   char tseper;
   int i, flag;
   MORAY_Object *obj;
   Vec tempv;
   double len;

   /* Walk through all the objects, defining them as we go */
   for (obj=objects,flag=0;obj!=NULL;obj=obj->next,flag=1) {
      /* Save the old transformation stack */
      lib_tx_push();

      /* Add any transformations to the current stack */
      MakeVector(obj->base_data.tx, obj->base_data.ty,
                 obj->base_data.tz, tempv)
      lib_tx_translate(tempv);
      lib_tx_rotate(2, degtorad(-obj->base_data.rz));
      lib_tx_rotate(1, degtorad(-obj->base_data.ry));
      lib_tx_rotate(0, degtorad(-obj->base_data.rx));
      MakeVector(obj->base_data.sx, obj->base_data.sy,
                 obj->base_data.sz, tempv)
      lib_tx_scale(tempv);

      /* First look for a few special cases */
      if (obj->type == MORAY_CAMERA_TYPE) {
         /* Do nothing, the camera must appear on the
            list of cameras */
         }
      else {
         /* For the general case, we declare the components of
            an object followed by the object itself. */
         indent(depth);
         if (depth == 0)
            message("object {\n");
         else if (flag)
            message(" %c object {\n", seper);
         else
            message("object {\n");

         /* Shape information */
         switch (obj->type) {
         case MORAY_NULL_TYPE:
            indent(depth+1);
            message("null_object\n");
            break;
         case MORAY_GROUP_TYPE:
            /* Same as a CSG union */
            Declare_Objects((MORAY_Object *)obj->obj_data, depth+1, '+');
            break;
         case MORAY_CUBE_TYPE:
            indent(depth+1);
            message("box <-1,-1,-1>,<1,1,1>\n");
            break;
         case MORAY_SPHERE_TYPE:
            indent(depth+1);
            message("sphere <0, 0, 0>, 1\n");
            break;
         case MORAY_CYLINDER_TYPE:
            dump_cylinder(depth+1);
            break;
         case MORAY_CONE_TYPE:
            dump_cone(depth+1);
            break;
         case MORAY_RSWEEP_TYPE:
            dump_lathe(depth+1, obj->obj_data);
            break;
         case MORAY_TSWEEP_TYPE:
         case MORAY_CSWEEP_TYPE:
            dump_glyph(depth+1, obj->obj_data);
            break;
         case MORAY_VOXEL_TYPE:
         case MORAY_BUMPMAP_TYPE:
            error("Reserved object type: %d in MORAY file\n", obj->type);
            break;
         case MORAY_CSG_TYPE:
            if (((MRY_CSG_Data *)obj->obj_data)->csg_type ==
                MORAY_CSG_INTERSECTION)
               tseper = '*';
            else if (((MRY_CSG_Data *)obj->obj_data)->csg_type ==
                     MORAY_CSG_DIFFERENCE)
               tseper = '-';
            else
               tseper = '+';
            Declare_Objects(((MRY_CSG_Data *)obj->obj_data)->objs,depth+1,
                            tseper);
            break;
         case MORAY_BEZIER_TYPE:
            dump_bezier(depth, (MRY_Bezier_Data *)obj->obj_data);
            break;
         case MORAY_HEIGHTFIELD_TYPE:
            dump_heightfield(depth+1, obj);
            break;
         case MORAY_TORUS_TYPE:
            dump_torus(depth+1, obj);
            break;
         case MORAY_PLANE_TYPE:
            dump_plane(depth+1);
            break;
         case MORAY_DISC_TYPE:
            dump_disc(depth+1, obj);
            break;
         case MORAY_PLIGHT_TYPE:
         case MORAY_SLIGHT_TYPE:
         case MORAY_ALIGHT_TYPE:
            dump_light(depth+1, obj);
            break;
         default:
            error("Unknown object type: %d in MORAY file\n", obj->type);
         }

         write_texture(depth+1, obj);
         if (test_transform_values(obj)) {
            indent(depth+1);
            write_transform_info(obj);
            message("\n");
            }

         indent(depth+1);
         message("}\n");
         }

      lib_tx_pop();
      }
}

static void
show_arguments(void)
{
message("Syntax: mryread filename\n");
}

int
main(int argc, char **argv)
{
   FILE *infile;
   int exit_code;
   MORAY_Model model;

#if defined( MAC )
   /* InitProfile(1000, 200); */
   argc = ccommand(&argv);
#endif

   if (argc < 2) {
      show_arguments();
      exit_code = 1;
      }
   else if ((infile = fopen(argv[1], "rb")) == NULL)
      error("Cannot open MORAY file: '%s'\n", argv[1]);
   else {
      Read_MORAY_File(infile, &model);
      Declare_Camera(&model);
      Declare_Textures(model.Texture_List);
      Declare_Objects(model.Object_List, 0, ' ');
      exit_code = 0;
      }

#if defined( MAC )
   /* freopen("profile.out", "w", stdout); */
#endif

   return exit_code;
}
