/* Code to read & store MORAY data files */
/* Platform/compiler specific include files */
#if defined( MAC )
#include <console.h>
#include <profile.h>
#elif defined( DOS386 )
#include <conio.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>
#include "mryread.h"

/******************* Internal Data ************************/

static int Selected_Object_Flag = 0;     /* None selected by default */
static char Selected_Object[MORAY_OBJ_NAME_LEN];

/******************* Internal processing ******************/
void
error(char *fmt, ...)
{
   va_list ap;

   fprintf(stderr, "ERROR: ");
   va_start(ap, fmt);
   vfprintf(stderr, fmt, ap);
   va_end(ap);

   exit(1);
}

void
message(char *fmt, ...)
{
   va_list ap;

   va_start(ap, fmt);
   vfprintf(stdout, fmt, ap);
   va_end(ap);

}

static byte
Read_Byte(FILE *infile)
{
   byte value;
   assert(sizeof(byte) == 1);

   if ((value = fgetc(infile)) == EOF)
      error("Truncated MORAY data file 1\n");
   return value;
}

static word
Read_Word(FILE *infile)
{
   ubyte low_byte, high_byte;
   assert(sizeof(word) == 2);

   if ((low_byte = fgetc(infile)) == EOF)
      error("Truncated MORAY data file 2\n");
   if ((high_byte = fgetc(infile)) == EOF)
      error("Truncated MORAY data file 3\n");
   return (word)((uword)low_byte + ((uword)high_byte << 8));
}


static long
Read_Long(FILE *infile)
{
   uword low_word, high_word;
   assert(sizeof(long) == 4);

   low_word = Read_Word(infile);
   high_word = Read_Word(infile);

   return (long)((long)low_word + ((ulong)high_word << 8));
}

static float
Read_Float(FILE *infile)
{
   float value;
#if defined( MAC )
   unsigned char temp;
#endif
   assert(sizeof(float) == 4);

   if (fread(&value, sizeof(float), 1, infile) != 1)
      error("Truncated MORAY data file 4\n");
#if defined( MAC )
   temp = ((unsigned char *)&value)[0];
   ((unsigned char *)&value)[0] = ((unsigned char *)&value)[3];
   ((unsigned char *)&value)[3] = temp;
   temp = ((unsigned char *)&value)[1];
   ((unsigned char *)&value)[1] = ((unsigned char *)&value)[2];
   ((unsigned char *)&value)[2] = temp;
#endif

   return value;
}

static double
Read_Double(FILE *infile)
{
   double value;
#if defined( MAC )
   unsigned char temp;
#endif

   if (fread(&value, sizeof(double), 1, infile) != 1)
      error("Truncated MORAY data file\n");

#if defined( MAC )
   temp = ((unsigned char *)&value)[0];
   ((unsigned char *)&value)[0] = ((unsigned char *)&value)[7];
   ((unsigned char *)&value)[7] = temp;
   temp = ((unsigned char *)&value)[1];
   ((unsigned char *)&value)[1] = ((unsigned char *)&value)[6];
   ((unsigned char *)&value)[6] = temp;
   temp = ((unsigned char *)&value)[2];
   ((unsigned char *)&value)[2] = ((unsigned char *)&value)[5];
   ((unsigned char *)&value)[5] = temp;
   temp = ((unsigned char *)&value)[3];
   ((unsigned char *)&value)[3] = ((unsigned char *)&value)[4];
   ((unsigned char *)&value)[4] = temp;
#endif

   return value;
}

/* cflag is set to 1 if it is a complete MORAY data file and to 0 if
   it only contains objects. */
static void
Read_MORAY_Header(FILE *infile, int *cflag, int *version)
{
   char header_block[6];

   if (fread(&header_block[0], 1, 6, infile) != 6)
      error("Truncated MORAY data file 6\n");
   if (header_block[0] != 'M' ||
       header_block[1] != 'D' ||
       header_block[2] != 'L')
      error("Invalid MORAY file\n");
   if (header_block[3] == 'X')
      /* Read complete MORAY data file */
      *cflag = 1;
   else if (header_block[3] == 'Y')
      /* Read incomplete MORAY data file */
      *cflag = 0;
   else
      error("Invalid MORAY data file");

   *version = (header_block[4] - '0') * 10 +
              (header_block[5] - '0');
#if DEBUG
message("%s model file, version: %d\n",
        (*cflag ? "complete" : "incomplete"),
        *version);
#endif
}

static void
Read_Tracer_Parameters(FILE *infile, int *xres, float *aspect, int *vislv)
{
   word wtemp;

   *xres   = Read_Word(infile); /* Image resolution in columns */
   *aspect = (float)Read_Word(infile) / 1000.0; /* Image aspect ratio */
   wtemp   = Read_Word(infile); /* Left edge percent should be 0 */
   wtemp   = Read_Word(infile); /* Right edge percent should be 100 */
   wtemp   = Read_Word(infile); /* Top edge percent should be 0 */
   wtemp   = Read_Word(infile); /* Bottom edge percent should be 100 */
   *vislv  = Read_Word(infile); /* Global visibility level */
   wtemp   = Read_Word(infile); /* Ambient red level */
   wtemp   = Read_Word(infile); /*         green     */
   wtemp   = Read_Word(infile); /*         blue      */
#if DEBUG
message("Res: %d, aspect: %f, Vis: %d\n", *xres, *aspect, *vislv);
#endif
}

/* See MORAY docs for this block - none of it is used here */
static void
Read_2D_View_Parameters(FILE *infile)
{
   byte btemp;
   double dtemp;

   dtemp = Read_Double(infile); /* X coord at edge of view */
   dtemp = Read_Double(infile); /* Width in X */
   dtemp = Read_Double(infile); /* Y coord at edge of view */
   dtemp = Read_Double(infile); /* Width in Y */
   dtemp = Read_Double(infile); /* Z coord at edge of view */
   dtemp = Read_Double(infile); /* Width in Z */
   btemp = Read_Byte(infile);   /* Draw grid flag */
   btemp = Read_Byte(infile);   /* Visible view flag */
   btemp = Read_Byte(infile);   /* Zoomed view flag */
   btemp = Read_Byte(infile);   /* Disabled grid flag */
}

/* Read the three orthogonal view parameters */
static void
Read_2D_Views(FILE *infile)
{
   Read_2D_View_Parameters(infile); /* XY (top) view */
   Read_2D_View_Parameters(infile); /* XZ (front) view */
   Read_2D_View_Parameters(infile); /* YZ (side) view */
}

static void
Read_Common_Data(FILE *infile, MORAY_Object *obj)
{
#if DEBUG
message("Read_Common_Data\n");
#endif
   obj->base_data.num = Read_Word(infile);
   if (fread(&obj->base_data.name[0], sizeof(char),
             MORAY_OBJ_NAME_LEN, infile) != MORAY_OBJ_NAME_LEN)
      error("Truncated MORAY data file 7\n");
#if DEBUG
message("Obj name: '%s'\n", &obj->base_data.name[0]);
#endif
   obj->base_data.sx = Read_Float(infile);
   obj->base_data.sy = Read_Float(infile);
   obj->base_data.sz = Read_Float(infile);
   obj->base_data.rx = Read_Float(infile);
   obj->base_data.ry = Read_Float(infile);
   obj->base_data.rz = Read_Float(infile);
   obj->base_data.tx = Read_Double(infile);
   obj->base_data.ty = Read_Double(infile);
   obj->base_data.tz = Read_Double(infile);
   obj->base_data.vislv = Read_Byte(infile);
   obj->base_data.relvs = Read_Byte(infile);
#if DEBUG
message("Num: %d, name: '%s'\n", obj->base_data.num,
        &obj->base_data.name[0]);
message("Scale <%f,%f,%f>\n",
        obj->base_data.sx,
        obj->base_data.sy,
        obj->base_data.sz);
message("Rotat <%f,%f,%f>\n",
        obj->base_data.rx,
        obj->base_data.ry,
        obj->base_data.rz);
message("Trans <%f,%f,%f>\n",
        obj->base_data.tx,
        obj->base_data.ty,
        obj->base_data.tz);
message("Vislv: %d, relvs: %d\n",
        (int)obj->base_data.vislv,
        (int)obj->base_data.relvs);
#endif
}

static MORAY_Object *
New_MORAY_Object(void)
{
   MORAY_Object *obj;
#if DEBUG
message("New_MORAY_Object\n");
#endif

   obj = malloc(sizeof(MORAY_Object));
   if (obj == NULL)
      error("Out of memory\n");

   obj->type = -1;

   obj->base_data.num = 0;
   obj->base_data.name[0] = '\0';
   obj->base_data.sx = 1.0;
   obj->base_data.sy = 1.0;
   obj->base_data.sz = 1.0;
   obj->base_data.rx = 0.0;
   obj->base_data.ry = 0.0;
   obj->base_data.rz = 0.0;
   obj->base_data.tx = 0.0;
   obj->base_data.ty = 0.0;
   obj->base_data.tz = 0.0;
   obj->base_data.vislv = 0;
   obj->base_data.relvs = 0;

   obj->texture = NULL;
   obj->bobj    = NULL;
   obj->next    = NULL;

   return obj;
}

static MORAY_Object *
Insert_New_Texture(MORAY_Model *model, MORAY_Object *texture)
{
   MORAY_Object *temp_tex;

   /* First see if the texture is already known */
   for (temp_tex = model->Texture_List;
        temp_tex != NULL;
        temp_tex = temp_tex->next)
      if (!strcmp(&texture->base_data.name[0],
                  &temp_tex->base_data.name[0]))
         break;

   /* Now return the pointer to the existing one or add it to
      the list if it isn't known. */
   if (temp_tex == NULL) {
      /* Never heard of it */
#if DEBUG
message("New Texture: '%s'\n", &texture->base_data.name[0]);
#endif
      texture->next = model->Texture_List;
      model->Texture_List = texture;

      return texture;
      }
   else {
      free(texture); /* We need to do more than this... */
#if DEBUG
message("Old Texture: '%s'\n", &temp_tex->base_data.name[0]);
#endif
      return temp_tex;
      }
}

static void
Delete_Texture(MORAY_Object *obj)
{
   MRY_Texture_Data  *base_texture;
   MRY_Imagemap_Data *imagemap;
   MRY_Checker_Data  *checker;
   int tex_type;

   /* Now read the standard texture information */
   base_texture = obj->obj_data;
   tex_type     = base_texture->texture_type;
   if (tex_type == MORAY_CHECKER_TEXTURE) {
      checker = base_texture->unique_data;
      free(checker);
      }
   else if (tex_type == MORAY_IMAGEMAP_TEXTURE) {
      imagemap = base_texture->unique_data;
      free(imagemap);
      }
   free(base_texture);
}

static void
Read_Base_Texture(FILE *infile, MORAY_Model *model, int tex_type,
                  MRY_Texture_Data *base_texture)
{
   int i;
   MRY_Imagemap_Data *imagemap;
   MRY_Checker_Data  *checker;
float t1, t2, t3;
int t4;
#if DEBUG
fpos_t filepos;
#endif

   /* Now read the standard texture information */
   base_texture->texture_type = tex_type;
   base_texture->name_len = Read_Byte(infile);
   if (base_texture->name_len > MORAY_OBJ_NAME_LEN-1)
      error("Texture name too long\n");
   for (i=0;i<base_texture->name_len;i++)
      base_texture->name[i] = Read_Byte(infile);
   base_texture->name[i] = '\0';
   base_texture->amb  = Read_Float(infile);
   base_texture->dif  = Read_Float(infile);
   base_texture->phg  = Read_Float(infile);
   base_texture->phgx = Read_Float(infile);
   base_texture->refl = Read_Float(infile);
   base_texture->trns = Read_Float(infile);
   base_texture->rfin = Read_Float(infile);
   base_texture->cldr = Read_Float(infile);
   base_texture->cldg = Read_Float(infile);
   base_texture->cldb = Read_Float(infile);
   if (model->Version > 25)
      base_texture->cldf = Read_Float(infile);
   else
      base_texture->cldf = 0.0;
   base_texture->bril = Read_Float(infile);
   base_texture->rdnm = Read_Float(infile);
   base_texture->spec = Read_Float(infile);
   base_texture->rghn = Read_Float(infile);
   if (model->Version > 25) {
      base_texture->turb = Read_Float(infile);
      base_texture->octv = Read_Float(infile);
      }
   else {
      base_texture->turb = 0.0;
      base_texture->octv = 1;
      }
   base_texture->clsr = Read_Float(infile);
   base_texture->clsg = Read_Float(infile);
   base_texture->clsb = Read_Float(infile);
   base_texture->metl = Read_Byte(infile);

#if DEBUG
fgetpos(infile, &filepos);
message("Done with texture at %lx/%ld\n", (long)filepos, (long)filepos);
#endif
#if DEBUG
message("Texture name: %s\n", &base_texture->name[0]);
message("   clr : <%g,%g,%g,%g>\n",
        (double)base_texture->cldr,
        (double)base_texture->cldg,
        (double)base_texture->cldb,
        (double)base_texture->cldf);
message("   amb : %g\n", (double)base_texture->amb);
message("   dif : %g\n", (double)base_texture->dif);
message("   bril: %g\n", (double)base_texture->bril);
message("   phg : %g\n", (double)base_texture->phg);
message("   phgx: %g\n", (double)base_texture->phgx);
message("   rdnm: %g\n", (double)base_texture->rdnm);
message("   spec: %g\n", (double)base_texture->spec);
message("   sclr: <%g,%g,%g>\n",
        (double)base_texture->clsr,
        (double)base_texture->clsg,
        (double)base_texture->clsb);
message("   rghn: %g\n", (double)base_texture->rghn);
message("   refl: %g\n", (double)base_texture->refl);
message("   trns: %g\n", (double)base_texture->trns);
message("   rfin: %g\n", (double)base_texture->rfin);
message("   metl: %d\n", (int)base_texture->metl);
message("   turb: %g\n", (double)t1);
message("   octv: %g\n", (double)t2);
#endif

   if (tex_type == MORAY_OPAQUE_TEXTURE) {
#if DEBUG
message("Opaque texture\n");
#endif
      base_texture->unique_data = NULL;
      }
   else if (tex_type == MORAY_WOOD_TEXTURE) {
#if DEBUG
message("Wood texture\n");
#endif
      base_texture->unique_data = NULL;
      }
   else if (tex_type == MORAY_GRADIENT_TEXTURE) {
#if DEBUG
message("Gradient texture\n");
#endif
      base_texture->unique_data = NULL;
      }
   else if (tex_type == MORAY_PREDEFINED_TEXTURE) {
#if DEBUG
message("Predefined texture\n");
#endif
      base_texture->unique_data = NULL;
      }
   else if (tex_type == MORAY_CHECKER_TEXTURE) {
#if DEBUG
message("Checker texture\n");
#endif
      checker = malloc(sizeof(MRY_Checker_Data));
      if (checker == NULL)
         error("Out of memory\n");
      checker->cl2r = Read_Float(infile);
      checker->cl2g = Read_Float(infile);
      checker->cl2b = Read_Float(infile);
      if (model->Version > 25)
         checker->cl2f = Read_Float(infile);
      else
         checker->cl2f = 0.0;
#if DEBUG
fgetpos(infile, &filepos);
message("Checker color <%g,%g,%g,%g>\n",
        (double)checker->cl2r, (double)checker->cl2g,
        (double)checker->cl2b, (double)checker->cl2f);
#endif
      base_texture->unique_data = checker;
      }
   else if (tex_type == MORAY_IMAGEMAP_TEXTURE) {
#if DEBUG
message("Image texture\n");
#endif
      imagemap = malloc(sizeof(MRY_Imagemap_Data));
      if (imagemap == NULL)
         error("Out of memory\n");
#if DEBUG
fgetpos(infile, &filepos);
message("Get imagemap name at %lx/%ld\n", (long)filepos, (long)filepos);
#endif
      imagemap->name_len = Read_Byte(infile);
      if (imagemap->name_len > MORAY_OBJ_NAME_LEN-1)
         error("image map name too long\n");
      for (i=0;i<imagemap->name_len;i++)
         imagemap->name[i] = Read_Byte(infile);
      imagemap->name[i] = '\0';
      imagemap->smth  = Read_Byte(infile);
      imagemap->tiled = Read_Byte(infile);
      if (model->Version > 24)
         imagemap->mapt  = Read_Byte(infile);
      else
         imagemap->mapt  = 0; /* Planar for versions before V25 */
      if (imagemap->mapt != MORAY_PLANAR_IMAGEMAP &&
          imagemap->mapt != MORAY_SPHERICAL_IMAGEMAP &&
          imagemap->mapt != MORAY_CYLINDRICAL_IMAGEMAP &&
          imagemap->mapt != MORAY_TOROIDAL_IMAGEMAP)
         error("Unknown mapping type: %d\n", imagemap->mapt);
      base_texture->unique_data = imagemap;
      }
   else
      error("Unknown texture type: %d\n", tex_type);
}

static MORAY_Object *
Read_Texture(FILE *infile, MORAY_Model *model)
{
   int tex_type, i;
   MRY_Texture_Data *base_texture;
   MORAY_Object *tex_obj;
#if DEBUG
fpos_t filepos;
fgetpos(infile, &filepos);
message("Read_Texture at %lx/%ld\n", (long)filepos, (long)filepos);
#endif

   tex_type = Read_Word(infile);

   if (tex_type != MORAY_NO_TEXTURE &&
       tex_type != MORAY_OPAQUE_TEXTURE &&
       tex_type != MORAY_CHECKER_TEXTURE &&
       tex_type != MORAY_WOOD_TEXTURE &&
       tex_type != MORAY_GRADIENT_TEXTURE &&
       tex_type != MORAY_IMAGEMAP_TEXTURE &&
       tex_type != MORAY_PREDEFINED_TEXTURE)
      error("Unknown texture type: %d\n", tex_type);

   if (tex_type == MORAY_NO_TEXTURE) {
#if DEBUG
message("No texture\n");
#endif
      return NULL;
      }
   else {
      tex_obj = New_MORAY_Object();
      tex_obj->type = MORAY_TEXTURE_TYPE;

      base_texture = malloc(sizeof(MRY_Texture_Data));
      if (base_texture == NULL)
         error("Out of memory\n");

      Read_Common_Data(infile, tex_obj);

      /* Read the actual texture information */
      Read_Base_Texture(infile, model, tex_type, base_texture);

      /* May fail if two global textures have the same name */
      tex_obj->obj_data = base_texture;
      tex_obj = Insert_New_Texture(model, tex_obj);

      return tex_obj;
      }
}

static void
Read_Camera(FILE *infile, MORAY_Object *obj)
{
   MRY_Camera_Data *camera;
   int t1;
#if DEBUG
   fpos_t filepos;

fgetpos(infile, &filepos);
message("Read_Camera at %lx/%ld\n", (long)filepos, (long)filepos);
#endif

   camera = malloc(sizeof(MRY_Camera_Data));
   if (camera == NULL)
      error("Out of memory\n");

   camera->eyepx = Read_Double(infile);
   camera->eyepy = Read_Double(infile);
   camera->eyepz = Read_Double(infile);
   camera->eyelx = Read_Double(infile);
   camera->eyely = Read_Double(infile);
   camera->eyelz = Read_Double(infile);
   camera->aprtr = Read_Double(infile);

#if DEBUG
message("From: <%g,%g,%g>, At: <%g,%g,%g>, Apr: %g\n",
        (double)camera->eyepx, (double)camera->eyepy, (double)camera->eyepz,
        (double)camera->eyelx, (double)camera->eyely, (double)camera->eyelz,
        (double)camera->aprtr);
#endif

   obj->obj_data = (void *)camera;
}

static void
Read_Point_Light(FILE *infile, MORAY_Object *obj)
{
   MRY_Point_Light_Data *light;
#if DEBUG
fpos_t filepos;
fgetpos(infile, &filepos);
message("Read_Point_Light at %lx/%ld\n", (long)filepos, (long)filepos);
#endif

   light = malloc(sizeof(MRY_Point_Light_Data));
   if (light == NULL)
      error("Out of memory\n");

   light->color[0] = Read_Float(infile);
   light->color[1] = Read_Float(infile);
   light->color[2] = Read_Float(infile);

#if DEBUG
message("Color: <%f,%f,%f>\n",
        light->color[0],
        light->color[1],
        light->color[2]);
#endif
   obj->obj_data = (void *)light;
}

static void
Read_Spot_Light(FILE *infile, MORAY_Object *obj)
{
   MRY_Spot_Light_Data *light;
#if DEBUG
message("Read_Spot_Light\n");
#endif

   light = malloc(sizeof(MRY_Spot_Light_Data));
   if (light == NULL)
      error("Out of memory\n");
   light->color[0]    = Read_Float(infile);
   light->color[1]    = Read_Float(infile);
   light->color[2]    = Read_Float(infile);
   light->falloff     = Read_Float(infile);
   light->radius      = Read_Float(infile);
   light->draw_length = Read_Float(infile);
   light->tightness   = Read_Word(infile);

#if DEBUG
message("Color: <%f,%f,%f>, falloff: %f, radius: %f, draw: %f, tight: %d\n",
        light->color[0], light->color[1], light->color[2],
        light->falloff, light->radius, light->draw_length,
        (int)light->tightness);
#endif
   obj->obj_data = (void *)light;
}

static void
Read_Area_Light(FILE *infile, MORAY_Object *obj)
{
   MRY_Area_Light_Data *light;
#if DEBUG
message("Read_Area_Light\n");
#endif

   light = malloc(sizeof(MRY_Area_Light_Data));
   if (light == NULL)
      error("Out of memory\n");
   light->color[0] = Read_Float(infile);
   light->color[1] = Read_Float(infile);
   light->color[2] = Read_Float(infile);
   light->xres     = Read_Word(infile);
   light->yres     = Read_Word(infile);
   light->adaptive = Read_Word(infile);
   light->jittr    = Read_Byte(infile);
   obj->obj_data = (void *)light;
}

/* Forward declaration for recursive object reads */
static MORAY_Object *Read_MORAY_Objects(FILE *infile, MORAY_Model *model,
                                        int *num_objs);
static MORAY_Object *Read_MORAY_Object(FILE *infile, MORAY_Model *model);
static void Read_Primitive(FILE *infile, MORAY_Model *model,
                           MORAY_Object *obj);

static void
Read_CSG(FILE *infile, MORAY_Model *model, MORAY_Object *obj)
{
   MRY_CSG_Data *csg_obj;
   int i;
#if DEBUG
fpos_t filepos;
fgetpos(infile, &filepos);
message("Read_CSG at %lx/%ld\n", (long)filepos, (long)filepos);
#endif

   csg_obj = malloc(sizeof(MRY_CSG_Data));
   if (csg_obj == NULL)
      error("Out of memory\n");
   csg_obj->objs = Read_MORAY_Objects(infile, model, &csg_obj->num_objs);
   csg_obj->csg_type = Read_Byte(infile);

   if (csg_obj->csg_type != MORAY_CSG_UNION &&
       csg_obj->csg_type != MORAY_CSG_INTERSECTION &&
       csg_obj->csg_type != MORAY_CSG_DIFFERENCE)
      error("Unknown csg type: %d\n", csg_obj->csg_type);

   obj->obj_data = (void *)csg_obj;
}

static void
Read_Bezier(FILE *infile, MORAY_Model *model, MORAY_Object *obj)
{
   MRY_Bezier_Data *bezier;
   MRY_Bezier_Vertices *vertices;
   MRY_Bezier_Patches *patches;
   MORAY_Object temp_obj;
   int i, j, t1, t2, t3, t4;
#if DEBUG
fpos_t filepos;
message("Read_Bezier\n");
#endif

   bezier = malloc(sizeof(MRY_Bezier_Data));
   if (bezier == NULL)
      error("Out of memory\n");
   bezier->stop   = Read_Word(infile);
   bezier->start  = Read_Word(infile);
   bezier->snapx  = Read_Double(infile);
   bezier->snapy  = Read_Double(infile);
   bezier->grid   = Read_Byte(infile);
   bezier->xyvxp  = Read_Double(infile);
   bezier->xyvyp  = Read_Double(infile);
   bezier->xyvdx  = Read_Double(infile);
   bezier->xyvdy  = Read_Double(infile);
   bezier->xzvxp  = Read_Double(infile);
   bezier->xzvzp  = Read_Double(infile);
   bezier->xzvdx  = Read_Double(infile);
   bezier->xzvdz  = Read_Double(infile);
   bezier->yzvyp  = Read_Double(infile);
   bezier->yzvzp  = Read_Double(infile);
   bezier->yzvdy  = Read_Double(infile);
   bezier->yzvdz  = Read_Double(infile);
   if (model->Version > 25) {
      bezier->usteps = Read_Word(infile);
      bezier->vsteps = Read_Word(infile);
      bezier->flatn  = Read_Double(infile);
      }
   else {
      bezier->usteps = 2;
      bezier->vsteps = 2;
      bezier->flatn  = 0.01;
      }
#if DEBUG
message("snapx: %g, snapy: %g, grid: %d\n",
        (double)bezier->snapx, (double)bezier->snapy, (int)bezier->grid);
message("xyv: %g, %g, %g\n",
        (double)bezier->xyvxp, (double)bezier->xyvyp,
        (double)bezier->xyvdx, (double)bezier->xyvdy);
message("xyv: %g, %g, %g\n",
        (double)bezier->xzvxp, (double)bezier->xzvzp,
        (double)bezier->xzvdx, (double)bezier->xzvdz);
message("xyv: %g, %g, %g\n",
        (double)bezier->yzvyp, (double)bezier->yzvzp,
        (double)bezier->yzvdy, (double)bezier->yzvdz);
message("usteps: %d, vsteps: %d, Flatness: %g\n",
        bezier->usteps, bezier->vsteps, (double)bezier->flatn);
#endif

   /* Read vertex list */
   j = bezier->stop - bezier->start;
#if DEBUG
message("%d points in bezier\n", j);
#endif
   bezier->points = malloc(j * sizeof(MRY_Bezier_Vertices));
   if (bezier->points == NULL)
      error("Out of memory\n");
   for (i=0;i<j;i++) {
      bezier->points[i].ptype = Read_Byte(infile);
      bezier->points[i].xpos  = Read_Double(infile);
      bezier->points[i].ypos  = Read_Double(infile);
      bezier->points[i].zpos  = Read_Double(infile);
      if (bezier->points[i].ptype == 1 ||
          bezier->points[i].ptype == 2 ||
          bezier->points[i].ptype == 4) {
         bezier->points[i].depnd1 = Read_Word(infile);
         bezier->points[i].depnd2 = Read_Word(infile);
#if DEBUG
message("depend: <%d,%d>\n", (int)bezier->points[i].depnd1,
        (int)bezier->points[i].depnd2);
#endif
         }
#if DEBUG
message("Point %d: <%g,%g,%g>, %d\n",
        (int)i,
        (double)bezier->points[i].xpos,
        (double)bezier->points[i].ypos,
        (double)bezier->points[i].zpos,
        (int)bezier->points[i].ptype);
#endif
      }

#if DEBUG
fgetpos(infile, &filepos);
message("Patches(%lx)\n", (long)filepos);
#endif

   bezier->ptchs = Read_Word(infile);
   bezier->xptch = Read_Word(infile);
   bezier->yptch = Read_Word(infile);

#if DEBUG
message("Patches: %d, xpatch: %d, ypatch: %d\n",
        (int)bezier->ptchs,
        (int)bezier->xptch,
        (int)bezier->yptch);
#endif

   bezier->patches = malloc(bezier->ptchs * sizeof(MRY_Bezier_Patches));
   if (bezier->patches == NULL)
      error("Out of memory\n");
   temp_obj.type = MORAY_NULL_TYPE;
   for (i=0;i<bezier->ptchs;i++) {
      /* The common data shouldn't be necessary here, but since it exists
         we will deal with it. */
      Read_Common_Data(infile, &temp_obj);
      Read_Primitive(infile, model, &temp_obj);

      /* Read the 16 control points of the sub patch */
      for (j=0;j<16;j++) {
         bezier->patches[i].vindices[j] = Read_Word(infile);
#if DEBUG
message("Patch[%d].Vindex[%d]: %d\n", (int)i, (int)j, bezier->patches[i].vindices[j]);
#endif
         }
      bezier->patches[i].subdiv = Read_Word(infile);
#if DEBUG
message("Subdiv[%d]: %d\n", (int)i, bezier->patches[i].subdiv);
#endif
      }

   obj->obj_data = (void *)bezier;
}

static void
Read_Torus(FILE *infile, MORAY_Object *obj)
{
   MRY_Torus_Data *torus;
#if DEBUG
message("Read_Torus\n");
#endif

   torus = malloc(sizeof(MRY_Torus_Data));
   if (torus == NULL)
      error("Out of memory\n");
   torus->inner_radius = Read_Double(infile);
   torus->outer_radius = Read_Double(infile);
#if DEBUG
message("Inner: %g, Outer: %g\n",
        (double)torus->inner_radius, (double)torus->outer_radius);
#endif
   obj->obj_data = (void *)torus;
}

static void
Read_Disc(FILE *infile, MORAY_Object *obj)
{
   MRY_Disc_Data *disc;
#if DEBUG
message("Read_Disc\n");
#endif

   disc = malloc(sizeof(MRY_Disc_Data));
   if (disc == NULL)
      error("Out of memory\n");
   disc->inner_radius = Read_Float(infile);
   disc->outer_radius = 1.0;
Read_Float(infile);
#if DEBUG
message("Inner: %g, Outer: %g\n",
        (double)disc->inner_radius, (double)disc->outer_radius);
#endif
   obj->obj_data = (void *)disc;
}

static void
Read_Heightfield(FILE *infile, MORAY_Object *obj)
{
   MRY_Heightfield_Data *hf;
   int i;
#if DEBUG
fpos_t filepos;
message("Read_Heightfield\n");
#endif

   hf = malloc(sizeof(MRY_Heightfield_Data));
   if (hf == NULL)
      error("Out of memory\n");
#if DEBUG
fgetpos(infile, &filepos);
message("Read height field name at %lx\n", (long)filepos);
#endif
   hf->name_len = Read_Byte(infile);
   if (hf->name_len > MORAY_IMAGE_NAME_LEN-1)
      error("Height field name too long\n");
   for (i=0;i<hf->name_len;i++)
      hf->name[i] = Read_Byte(infile);
   hf->name[i] = '\0';
   hf->smooth      = Read_Byte(infile);
   hf->boxit       = Read_Byte(infile);
   hf->water_level = Read_Word(infile);
   obj->obj_data = (void *)hf;
}

static void
Read_Sweep(FILE *infile, MORAY_Model *model, MORAY_Object *obj)
{
   int i;
   MRY_Sweep_Point *points;
   MRY_Sweep_Data *sweep;
#if DEBUG
message("Read_Sweep\n");
#endif

   sweep = malloc(sizeof(MRY_Sweep_Data));
   if (sweep == NULL)
      error("Out of memory\n");
   sweep->numseg = Read_Word(infile);
   points = malloc(sweep->numseg * sizeof(MRY_Sweep_Point));
   if (points == NULL)
      error("Out of memory\n");

#if DEBUG
message("%d nodes in sweep\n", sweep->numseg);
#endif
   /* Read the nodes that make up the sweep */
   for (i=0;i<sweep->numseg;i++) {
      points[i].xpos = Read_Double(infile);
      points[i].ypos = Read_Double(infile);
      points[i].angl1 = Read_Double(infile);
      points[i].angl2 = Read_Double(infile);
#if DEBUG
      message("P[%d] <%g,%g>, a1: %g, a2: %g\n",
             i, (double)points[i].xpos, (double)points[i].ypos,
             (double)points[i].angl1, (double)points[i].angl2);
#endif
      }

   /* Now read the rest of the data structure */
   sweep->subdiv = Read_Word(infile);
   sweep->viewx  = Read_Double(infile);
   sweep->viewy  = Read_Double(infile);
   sweep->deltax = Read_Double(infile);
   if (model->Version > 25)
      sweep->expdiv = Read_Word(infile);
   else
      sweep->expdiv = 1;

#if DEBUG
   message("Subdiv: %d, View: <%g,%g>, Deltax: %g, Expdiv: %d\n",
          sweep->subdiv, (double)sweep->viewx, (double)sweep->viewy,
          (double)sweep->deltax, (int)sweep->expdiv);
#endif

   sweep->points = points;
   obj->obj_data = sweep;
}

static void
Read_Primitive(FILE *infile, MORAY_Model *model, MORAY_Object *obj)
{
   int i;
#if DEBUG
fpos_t filepos;
fgetpos(infile, &filepos);
message("Read_Primitive at %lx/%ld\n", (long)filepos, (long)filepos);
#endif

   /* All objects have texture data (even if it's just a flag
      that says there's no texture).  Note that we may actually
      have to do something else here, depending on wether or not
      a texture in an object definition contains common data.  */
   obj->texture = Read_Texture(infile, model);

   /* Check for a bounding box */
   i = Read_Byte(infile);
   if (i == 0) {
      obj->bobj = NULL;
      }
   else if (i == 1) {
      /* There is a bound expected for this primitive */
      obj->bobj = Read_MORAY_Object(infile, model);
      }
   else
      error("Bad bounding byte: %d\n", i);

   /* Does this object cast a shadow? */
   if (model->Version > 25)
      obj->noshdw = Read_Byte(infile);
   else
      obj->noshdw = 0;

#if DEBUG
fgetpos(infile, &filepos);
message("read object data for %s(%p) at %lx/%ld\n",
        &obj->base_data.name[0], obj, (long)filepos, (long)filepos);
#endif
   switch (obj->type) {
      case MORAY_NULL_TYPE:
#if DEBUG
message("Null object\n");
#endif
         break;
      case MORAY_GROUP_TYPE:
#if DEBUG
message("Read Group\n");
#endif
         obj->obj_data = (void *)Read_MORAY_Objects(infile, model, &i);
         break;
      /* The following primitives (cube, sphere, cylinder, bound,
         cone, and plane) add no additional data. */
      case MORAY_CUBE_TYPE:
      case MORAY_SPHERE_TYPE:
      case MORAY_CYLINDER_TYPE:
      case MORAY_BOUND_TYPE:
      case MORAY_CONE_TYPE:
      case MORAY_PLANE_TYPE:
         break;
      case MORAY_TSWEEP_TYPE:
      case MORAY_RSWEEP_TYPE:
      case MORAY_CSWEEP_TYPE:
         Read_Sweep(infile, model, obj);
         break;
      case MORAY_VOXEL_TYPE:
      case MORAY_BUMPMAP_TYPE:
         error("Reserved object type: %d in MORAY file\n", obj->type);
         break;
      case MORAY_CSG_TYPE:
         Read_CSG(infile, model, obj);
         break;
      case MORAY_BEZIER_TYPE:
         Read_Bezier(infile, model, obj);
         break;
      case MORAY_HEIGHTFIELD_TYPE:
         Read_Heightfield(infile, obj);
         break;
      case MORAY_TORUS_TYPE:
         Read_Torus(infile, obj);
         break;
      case MORAY_PLIGHT_TYPE:
         Read_Point_Light(infile, obj);
         break;
      case MORAY_SLIGHT_TYPE:
         Read_Spot_Light(infile, obj);
         break;
      case MORAY_ALIGHT_TYPE:
         Read_Area_Light(infile, obj);
         break;
      case MORAY_DISC_TYPE:
         Read_Disc(infile, obj);
         break;
      case MORAY_CAMERA_TYPE:
         Read_Camera(infile, obj);
         break;
      default:
         error("Unknown object type: %d in MORAY file\n", obj->type);
      }

#if DEBUG
fgetpos(infile, &filepos);
message("Done with prim at %lx/%ld\n", (long)filepos, (long)filepos);
#endif
}

/* Read a single object from the MORAY data file */
static MORAY_Object *
Read_MORAY_Object(FILE *infile, MORAY_Model *model)
{
   MORAY_Object *obj;
#if DEBUG
message("Read_MORAY_Object\n");
#endif

   obj = New_MORAY_Object();
   obj->type = Read_Byte(infile);
#if DEBUG
message("type: %d\n", (int)obj->type);
#endif

   Read_Common_Data(infile, obj);

   if (obj->type == MORAY_TEXTURE_TYPE)
      obj->texture = Read_Texture(infile, model);
   else
      Read_Primitive(infile, model, obj);

   return obj;
}

/* Read in a set of objects, first comes the count of objects, then
   the data for the objects themselves */
static MORAY_Object *
Read_MORAY_Objects(FILE *infile, MORAY_Model *model, int *obj_cnt)
{
   MORAY_Object *obj, *objs, *last_obj;
   int i;
#if DEBUG
fpos_t filepos;
fgetpos(infile, &filepos);
message("Read_MORAY_Objects at %ld\n", (long)filepos);
#endif

   i = *obj_cnt = Read_Word(infile);
   for (objs=NULL,last_obj=objs;i>0;i--) {
#if DEBUG
fgetpos(infile, &filepos);
message("Get obj at %lx\n", (long)filepos);
#endif
      obj = Read_MORAY_Object(infile, model);
      if (objs == NULL) {
         objs = obj;
         last_obj = objs;
         }
      else {
         last_obj->next = obj;
         last_obj = obj;
         }
      }
   return objs;
}

/* Read in a set of objects, first comes the count of objects, then
   the data for the objects themselves */
static MORAY_Object *
Read_MORAY_Cameras(FILE *infile, MORAY_Model *model)
{
   MORAY_Object *obj, *objs, *root_obj, *last_obj;
   int i, j, num_cams, len;
#if DEBUG
fpos_t filepos;
fgetpos(infile, &filepos);
message("Read_MORAY_Cameras at %ld\n", (long)filepos);
#endif

   root_obj = NULL;
   last_obj = NULL;
   num_cams = Read_Byte(infile);
   for (i=num_cams;i>0;i--) {
      obj = New_MORAY_Object();
      obj->type = MORAY_CAMERA_TYPE;
      Read_Common_Data(infile, obj);
      Read_Primitive(infile, model, obj);
      if (last_obj == NULL) {
         obj->next = NULL;
         last_obj  = obj;
         root_obj  = obj;
         }
      else {
         obj->next = NULL;
         last_obj->next = obj;
         last_obj = obj;
         }
      if (model->Version > 25) {
         /* Read in posdep: Name of object that camera/from is relative to */
         len = Read_Byte(infile);
         if (len > MORAY_OBJ_NAME_LEN-1)
            error("camera/from name too long\n");
         for (j=0;j<len;j++)
            (void)Read_Byte(infile);
         /* Read in lkatdep: Name of object that camera/at is relative to */
         len = Read_Byte(infile);
         if (len > MORAY_OBJ_NAME_LEN-1)
            error("camera/at name too long\n");
         for (j=0;j<len;j++)
            (void)Read_Byte(infile);
         }
      }

   return root_obj;
}

static void
Read_Selected_Object(FILE *infile, MORAY_Model *model)
{
   int i, len;
#if DEBUG
fpos_t filepos;
fgetpos(infile, &filepos);
message("Read_Selected_Object at %lx/%ld\n", (long)filepos, (long)filepos);
#endif

   if (model->Version > 24) {
      Selected_Object_Flag = Read_Byte(infile);
      /* The type of object selected is:
            0 - No object selected
            1 - A camera is selected
            2 - A texture is selected */

      /* Get the name of the selected object */
      if (Selected_Object_Flag) {
         len = Read_Byte(infile);
         for (i=0;i<len;i++)
            Selected_Object[i] = Read_Byte(infile);
         Selected_Object[i] = '\0';
         }
      else
         Selected_Object[0] = '\0';
      }
   else {
      /* In versions before 25 there is no selected object */
      Selected_Object_Flag = 0;
      Selected_Object[0] = '\0';
      }
#if DEBUG
if (Selected_Object_Flag == 0)
   message("No selected object\n");
else if (Selected_Object_Flag == 1)
   message("Selected camera: %s\n", Selected_Object);
else
   message("Selected object: %s\n", Selected_Object);
#endif
}

/* Read an entire MORAY file */
void
Read_MORAY_File(FILE *infile, MORAY_Model *model)
{
   int cflag;
   int vislv;

   /* Initialize the model */
   model->Version      = 0;
   model->Aspect       = 1.0;
   model->XResolution  = 256;
   model->Object_Count = 0;
   model->Object_List  = NULL;
   model->Texture_List = NULL;
   model->Camera_List  = NULL;

   /* Read the header from the file to determine what version
      Moray file this is, and whether it is a single model or
      an entire scene. */
   Read_MORAY_Header(infile, &cflag, &model->Version);

   if (cflag) {
      /* Read a complete MORAY file */
      Read_Tracer_Parameters(infile, &model->XResolution,
                             &model->Aspect, &vislv);
      Read_2D_Views(infile);
      model->Object_Count = Read_Word(infile);
      model->Object_List = Read_MORAY_Objects(infile, model,
                                              &model->Object_Count);
      model->Camera_List = Read_MORAY_Cameras(infile, model);
      /* No need to read the selected object, we are going
         to manipulate everything. */
      Read_Selected_Object(infile, model);
      }
   else {
      /* Read a single model */
      model->Object_Count = Read_Word(infile);
      if (model->Object_Count == 1)
         model->Object_List = Read_MORAY_Object(infile, model);
      else {
         model->Object_List = NULL;
         error("Object count should be 1 (is %d) in single object file\n",
               model->Object_Count);
         }
      }
}
